from holon.principles import Principle

level0 = Principle(
    name="RMM Level 0",
    reference="https://martinfowler.com/articles/richardsonMaturityModel.html#level0",
)
level1 = Principle(
    name="RMM Level 1",
    reference="https://martinfowler.com/articles/richardsonMaturityModel.html#level1",
)
level2 = Principle(
    name="RMM Level 2",
    reference="https://martinfowler.com/articles/richardsonMaturityModel.html#level2",
)
level3 = Principle(
    name="RMM Level 3",
    reference="https://martinfowler.com/articles/richardsonMaturityModel.html#level3",
)



"""
A majority of this file is copied from FastAPI.
https://github.com/tiangolo/fastapi
"""

from enum import Enum
from typing import Any, Callable, Dict, Iterable, List, Literal, Optional, Union

from holon.base import autocomplete
from pydantic import AnyUrl, BaseModel, Field


@autocomplete
class Contact(BaseModel):
    name: Optional[str] = None
    url: Optional[AnyUrl] = None
    email: Optional[str] = None


@autocomplete
class License(BaseModel):
    name: str
    url: Optional[AnyUrl] = None


@autocomplete
class Info(BaseModel):
    title: str
    version: str
    description: Optional[str] = None
    termsOfService: Optional[str] = None
    contact: Optional[Contact] = None
    license: Optional[License] = None


@autocomplete
class ServerVariable(BaseModel):
    default: str
    enum: Optional[List[str]] = None
    description: Optional[str] = None


@autocomplete
class Server(BaseModel):
    url: Union[AnyUrl, str]
    description: Optional[str] = None
    variables: Optional[Dict[str, ServerVariable]] = None


class Reference(BaseModel):
    ref: str = Field(..., alias="$ref")


@autocomplete
class Discriminator(BaseModel):
    propertyName: str
    mapping: Optional[Dict[str, str]] = None


@autocomplete
class XML(BaseModel):
    name: Optional[str] = None
    namespace: Optional[str] = None
    prefix: Optional[str] = None
    attribute: Optional[bool] = None
    wrapped: Optional[bool] = None


@autocomplete
class ExternalDocumentation(BaseModel):
    url: AnyUrl
    description: Optional[str] = None


@autocomplete
class SchemaBase(BaseModel):
    ref: Optional[str] = Field(None, alias="$ref")
    title: Optional[str] = None
    multipleOf: Optional[float] = None
    maximum: Optional[float] = None
    exclusiveMaximum: Optional[float] = None
    minimum: Optional[float] = None
    exclusiveMinimum: Optional[float] = None
    maxLength: Optional[int] = Field(None, gte=0)
    minLength: Optional[int] = Field(None, gte=0)
    pattern: Optional[str] = None
    maxItems: Optional[int] = Field(None, gte=0)
    minItems: Optional[int] = Field(None, gte=0)
    uniqueItems: Optional[bool] = None
    maxProperties: Optional[int] = Field(None, gte=0)
    minProperties: Optional[int] = Field(None, gte=0)
    required: Optional[List[str]] = None
    enum: Optional[List[Any]] = None
    type: Optional[str] = None
    allOf: Optional[List[Any]] = None
    oneOf: Optional[List[Any]] = None
    anyOf: Optional[List[Any]] = None
    not_: Optional[Any] = Field(None, alias="not")
    items: Optional[Any] = None
    properties: Optional[Dict[str, Any]] = None
    additionalProperties: Optional[Union[Dict[str, Any], bool]] = None
    description: Optional[str] = None
    format: Optional[str] = None
    default: Optional[Any] = None
    nullable: Optional[bool] = None
    discriminator: Optional[Discriminator] = None
    readOnly: Optional[bool] = None
    writeOnly: Optional[bool] = None
    xml: Optional[XML] = None
    externalDocs: Optional[ExternalDocumentation] = None
    example: Optional[Any] = None
    deprecated: Optional[bool] = None


@autocomplete
class Schema(SchemaBase):
    allOf: Optional[List[SchemaBase]] = None
    oneOf: Optional[List[SchemaBase]] = None
    anyOf: Optional[List[SchemaBase]] = None
    not_: Optional[SchemaBase] = Field(None, alias="not")
    items: Optional[SchemaBase] = None
    properties: Optional[Dict[str, SchemaBase]] = None
    additionalProperties: Optional[Union[Dict[str, Any], bool]] = None


@autocomplete
class Example(BaseModel):
    summary: Optional[str] = None
    description: Optional[str] = None
    value: Optional[Any] = None
    externalValue: Optional[AnyUrl] = None


@autocomplete
class ParameterInType(Enum):
    query = "query"
    header = "header"
    path = "path"
    cookie = "cookie"


@autocomplete
class Encoding(BaseModel):
    contentType: Optional[str] = None
    # Workaround OpenAPI recursive reference, using Any
    headers: Optional[Dict[str, Union[Any, Reference]]] = None
    style: Optional[str] = None
    explode: Optional[bool] = None
    allowReserved: Optional[bool] = None


@autocomplete
class MediaType(BaseModel):
    schema_: Optional[Union[Schema, Reference]] = Field(None, alias="schema")
    example: Optional[Any] = None
    examples: Optional[Dict[str, Union[Example, Reference]]] = None
    encoding: Optional[Dict[str, Encoding]] = None


@autocomplete
class ParameterBase(BaseModel):
    description: Optional[str] = None
    required: Optional[bool] = None
    deprecated: Optional[bool] = None
    # Serialization rules for simple scenarios
    style: Optional[str] = None
    explode: Optional[bool] = None
    allowReserved: Optional[bool] = None
    schema_: Optional[Union[Schema, Reference]] = Field(None, alias="schema")
    example: Optional[Any] = None
    examples: Optional[Dict[str, Union[Example, Reference]]] = None
    # Serialization rules for more complex scenarios
    content: Optional[Dict[str, MediaType]] = None


@autocomplete
class Parameter(ParameterBase):
    name: str  # type: ignore
    in_: ParameterInType = Field(..., alias="in")


@autocomplete
class Header(ParameterBase):
    pass


# Workaround OpenAPI recursive reference
@autocomplete
class EncodingWithHeaders(Encoding):
    headers: Optional[Dict[str, Union[Header, Reference]]] = None


@autocomplete
class RequestBody(BaseModel):
    description: Optional[str] = None
    content: Dict[str, MediaType]  # type: ignore
    required: Optional[bool] = None


@autocomplete
class Link(BaseModel):
    operationRef: Optional[str] = None
    operationId: Optional[str] = None
    parameters: Optional[Dict[str, Union[Any, str]]] = None
    requestBody: Optional[Union[Any, str]] = None
    description: Optional[str] = None
    server: Optional[Server] = None


@autocomplete
class Response(BaseModel):
    description: str
    headers: Optional[Dict[str, Union[Header, Reference]]] = None
    content: Optional[Dict[str, MediaType]] = None
    links: Optional[Dict[str, Union[Link, Reference]]] = None


@autocomplete
class Operation(BaseModel):
    tags: Optional[List[str]] = None
    summary: Optional[str] = None
    description: Optional[str] = None
    externalDocs: Optional[ExternalDocumentation] = None
    operationId: Optional[str] = None
    parameters: Optional[List[Union[Parameter, Reference]]] = None
    requestBody: Optional[Union[RequestBody, Reference]] = None
    responses: Dict[str, Response]  # type: ignore
    # Workaround OpenAPI recursive reference
    callbacks: Optional[Dict[str, Union[Dict[str, Any], Reference]]] = None
    deprecated: Optional[bool] = None
    security: Optional[List[Dict[str, List[str]]]] = None
    servers: Optional[List[Server]] = None


@autocomplete
class PathItem(BaseModel):
    ref: Optional[str] = Field(None, alias="$ref")
    summary: Optional[str] = None
    description: Optional[str] = None
    get: Optional[Operation] = None
    put: Optional[Operation] = None
    post: Optional[Operation] = None
    delete: Optional[Operation] = None
    options: Optional[Operation] = None
    head: Optional[Operation] = None
    patch: Optional[Operation] = None
    trace: Optional[Operation] = None
    servers: Optional[List[Server]] = None
    parameters: Optional[List[Union[Parameter, Reference]]] = None


# Workaround OpenAPI recursive reference
@autocomplete
class OperationWithCallbacks(BaseModel):
    callbacks: Optional[Dict[str, Union[Dict[str, PathItem], Reference]]] = None


@autocomplete
class SecuritySchemeType(Enum):
    apiKey = "apiKey"
    http = "http"
    oauth2 = "oauth2"
    openIdConnect = "openIdConnect"


@autocomplete
class SecurityBase(BaseModel):
    type_: SecuritySchemeType = Field(..., alias="type")
    description: Optional[str] = None


@autocomplete
class APIKeyIn(Enum):
    query = "query"
    header = "header"
    cookie = "cookie"


@autocomplete
class APIKey(SecurityBase):
    type_ = Field(SecuritySchemeType.apiKey, alias="type")
    in_: APIKeyIn = Field(..., alias="in")
    name: str  # type: ignore


@autocomplete
class HTTPBase(SecurityBase):
    type_ = Field(SecuritySchemeType.http, alias="type")
    scheme: str  # type: ignore


@autocomplete
class HTTPBearer(HTTPBase):  # type: ignore
    scheme = "bearer"
    bearerFormat: Optional[str] = None


@autocomplete
class OAuthFlow(BaseModel):
    refreshUrl: Optional[str] = None
    scopes: Dict[str, str] = {}


@autocomplete
class OAuthFlowImplicit(OAuthFlow):
    authorizationUrl: str  # type: ignore


@autocomplete
class OAuthFlowPassword(OAuthFlow):
    tokenUrl: str  # type: ignore


@autocomplete
class OAuthFlowClientCredentials(OAuthFlow):
    tokenUrl: str  # type: ignore


@autocomplete
class OAuthFlowAuthorizationCode(OAuthFlow):
    authorizationUrl: str  # type: ignore
    tokenUrl: str  # type: ignore


@autocomplete
class OAuthFlows(BaseModel):
    implicit: Optional[OAuthFlowImplicit] = None
    password: Optional[OAuthFlowPassword] = None
    clientCredentials: Optional[OAuthFlowClientCredentials] = None
    authorizationCode: Optional[OAuthFlowAuthorizationCode] = None


@autocomplete
class OAuth2(SecurityBase):
    type_ = Field(SecuritySchemeType.oauth2, alias="type")
    flows: OAuthFlows  # type: ignore


@autocomplete
class OpenIdConnect(SecurityBase):
    type_ = Field(SecuritySchemeType.openIdConnect, alias="type")
    openIdConnectUrl: str  # type: ignore


SecurityScheme = Union[APIKey, HTTPBase, OAuth2, OpenIdConnect, HTTPBearer]


@autocomplete
class Components(BaseModel):
    schemas: Optional[Dict[str, Union[Schema, Reference]]] = None
    responses: Optional[Dict[str, Union[Response, Reference]]] = None
    parameters: Optional[Dict[str, Union[Parameter, Reference]]] = None
    examples: Optional[Dict[str, Union[Example, Reference]]] = None
    requestBodies: Optional[Dict[str, Union[RequestBody, Reference]]] = None
    headers: Optional[Dict[str, Union[Header, Reference]]] = None
    securitySchemes: Optional[Dict[str, Union[SecurityScheme, Reference]]] = None
    links: Optional[Dict[str, Union[Link, Reference]]] = None
    callbacks: Optional[Dict[str, Union[Dict[str, PathItem], Reference]]] = None


@autocomplete
class Tag(BaseModel):
    name: str
    description: Optional[str] = None
    externalDocs: Optional[ExternalDocumentation] = None


@autocomplete
class OpenAPI(BaseModel):
    openapi: Literal["3.0.0"] = "3.0.0"
    info: Info  # type: ignore
    servers: Optional[List[Server]] = None
    paths: Dict[str, PathItem]  # type: ignore
    components: Optional[Components] = None
    security: Optional[List[Dict[str, List[str]]]] = None
    tags: Optional[List[Tag]] = None
    externalDocs: Optional[ExternalDocumentation] = None



from enum import Enum
from holon.base import autocomplete, Convention, ConventionRule


@autocomplete
class URIElement(Enum):
    RESOURCE_NAME = "Resource Name"
    TEMPLATES = "Templates"


@autocomplete
class URIConventionRule(ConventionRule[URIElement]):
    pass


nest_resources = Convention(name="Nest Resources")
include_version = Convention(name="Include Version")
uuid_as_id = Convention(name="UUID as ID")



from enum import Enum
from typing import Any, Dict, ForwardRef, List, Literal, Optional, Union
from pydantic import BaseModel, root_validator
from holon.base import autocomplete


class Format(Enum):
    TEXT = "text"
    HTML = "html"
    ASCIIDOC = "asciidoc"
    MARKDOWN = "markdown"


class DescriptorType(Enum):
    SEMANTIC = "semantic"
    SAFE = "safe"
    UNSAFE = "unsafe"
    IDEMPOTENT = "idempotent"


@autocomplete
class Doc(BaseModel):
    text: str
    href: Optional[str] = None
    format: Format = Format.TEXT


@autocomplete
class Ext(BaseModel):
    href: str


@autocomplete
class Link(BaseModel):
    rel: str
    href: str


@autocomplete
class Descriptor(BaseModel):
    """
    A 'descriptor' element defines the semantics of specific data
    elements or state transitions that MAY exist in an associated
    representation.
    """

    id: Optional[str] = None
    name: Optional[str] = None
    doc: Optional[Doc] = None
    rel: Optional[str] = None
    rt: Optional[str] = None
    type: DescriptorType = DescriptorType.SEMANTIC
    descriptors: List["Descriptor"] = []
    links: List[Link] = []

    @root_validator
    def check_id_name(cls, value: Dict[str, Any]) -> Dict[str, Any]:
        if value["id"] is None and value["name"] is None:
            raise ValueError("You must set either the id or name of a descriptor")
        return value


Descriptor.update_forward_refs()


@autocomplete
class ALPS(BaseModel):
    version: Literal["1.0"] = "1.0"
    title: Optional[str] = None
    doc: Optional[Doc] = None
    links: List[Link] = []
    descriptors: List[Descriptor] = []



from enum import Enum
from typing import Dict, ForwardRef, List, Literal, Optional, Union
from holon.base import autocomplete, Definition, DefinitionRule


@autocomplete
class Principle(Definition):
    pass


@autocomplete
class PrincipleRule(DefinitionRule):
    principle: Principle


robustness = Principle(
    name="Robustness Principle",
    reference="https://en.wikipedia.org/wiki/Robustness_principle",
)

# TODO: Further reading on downside
# https://tools.ietf.org/id/draft-thomson-postel-was-wrong-03.html

yagni = Principle(name="YAGNI", reference="https://martinfowler.com/bliki/Yagni.html")

# TODO: find good link for these
# TODO: move to practice
design_first = Principle(name="Design First")
code_first = Principle(name="Code First")
api_first = Principle(name="API First")



from enum import Enum
from holon.base import Convention

upper_case = Convention(name="Upper Case")
lower_case = Convention(name="Lower Case")
snake_cake = Convention(name="Snake Case")
camel_case = Convention(name="Camel Case")
kebab_case = Convention(name="Kebab Case")
screaming_snake_case = Convention(name="Screaming Snake Case")



from enum import Enum
from holon.base import autocomplete, Convention, ConventionRule


class JSONElement(Enum):
    ITEM = "ITEM"
    PROPERTY = "Property"
    VALUE = "Value"
    OBJECT = "Object"
    ARRAY = "Array"
    STRING = "String"
    NUMBER = "Number"
    NULL = "Null"


@autocomplete
class JSONConventionRule(ConventionRule[JSONElement]):
    pass


allow_null = Convention(name="Allow NULL")



from holon.protocols.http import Header

REFERENCE = "https://www.iana.org/go/rfc7232"


etag = Header(name="Etag", reference=REFERENCE)
last_modified = Header(name="Last-Modified", reference=REFERENCE)
if_match = Header(name="If-Match", reference=REFERENCE)
if_modified_since = Header(name="If-Modified-Since", reference=REFERENCE)
if_unmodified_since = Header(name="If-Unmodified-Since", reference=REFERENCE)
if_none_match = Header(name="If-None-Match", reference=REFERENCE)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['holon',
 'holon.conventions',
 'holon.formats',
 'holon.ietf',
 'holon.models',
 'holon.protocols']

package_data = \
{'': ['*']}

install_requires = \
['pydantic>=1.7.2,<2.0.0']

setup_kwargs = {
    'name': 'holon',
    'version': '0.1.1',
    'description': 'Experimental API Design Dystem library',
    'long_description': '# Holon\n\nTool for building API Design Systems.\n\n**Currently in experimental mode. Interface will change.**\n',
    'author': 'Stephen Mizell',
    'author_email': 'smizell@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/smizell/holon',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>3.6',
}


setup(**setup_kwargs)



from enum import Enum

# from holon.conventions import Convention
from holon.protocols.http import HTTPMessage, HTTPMethod, HTTPRuleSet
from holon.principles import Principle, PrincipleRule
from holon.base import autocomplete, BaseModel
from holon.ietf.rfc2119 import Keyword
from typing import List, Optional


# @autocomplete
# class PrincipleRule(BaseModel):
#     keyword: Keyword
#     principle: Principle


# @autocomplete
# class HeaderRule(BaseModel):
#     keyword: Keyword
#     header: Header


# @autocomplete
# class MediaTypeRule(BaseModel):
#     keyword: Keyword
#     media_type: MediaType


# @autocomplete
# class ConventionRule(BaseModel):
#     keyword: Keyword
#     convention: Convention


# @autocomplete
# class HTTPConventionRule(BaseModel):
#     keyword: Keyword
#     convention: Convention
#     element: Optional[http.HTTPElement] = None
#     method: Optiona[http.HTTPMethod] = None
#     message: Optional[http.HTTPMessage] = None


# @autocomplete
# class HTTPRuleSet(BaseModel):
#     headers: Optional[List[HeaderRule]] = []
#     media_types: Optional[List[MediaTypeRule]] = []
#     http_conventions: Optional[List[HTTPConventionRule]] = []


@autocomplete
class APIDesignSystem(BaseModel):
    principles: Optional[List[PrincipleRule]] = []
    http_rules: Optional[HTTPRuleSet] = None



from holon.principles import Principle

SLIDE_URL = "http://amundsen.com/talks/2016-11-apistrat-wadm/2016-11-apistrat-wadm.pdf"

database_centric = Principle(name="Database Centric", reference=SLIDE_URL)
object_centric = Principle(name="Object Centric", reference=SLIDE_URL)
resource_centric = Principle(name="Resource Centric", reference=SLIDE_URL)
affordance_centric = Principle(name="Affordance Centric", reference=SLIDE_URL)



from enum import Enum
from pydantic import BaseModel
from typing import TYPE_CHECKING, Generic, List, Optional, TypeVar, Union
from holon.ietf.rfc2119 import Keyword

if TYPE_CHECKING:
    from dataclasses import dataclass as autocomplete
else:

    def autocomplete(model):
        return model


T = TypeVar("T")
S = TypeVar("S")


@autocomplete
class Definition(BaseModel):
    name: str
    reference: Optional[str] = None


@autocomplete
class DefinitionRule(BaseModel):
    requirement: Keyword


@autocomplete
class Convention(Definition):
    pass


@autocomplete
class ConventionRule(DefinitionRule, Generic[T]):
    requirement: Keyword
    convention: Convention
    subject: Union[T, List[T]]


@autocomplete
class Ruleset(BaseModel):
    pass



from enum import Enum
from http import HTTPStatus
from typing import List, Optional, Union
from holon.base import autocomplete, ConventionRule, Definition, DefinitionRule, Ruleset


class HTTPMethod(Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"
    HEAD = "HEAD"
    OPTIONS = "OPTIONS"
    TRACE = "TRACE"


class HTTPMessage(Enum):
    REQUEST = "request"
    RESPONSE = "response"


class HTTPHeader(Enum):
    NAME = "Name"
    VALUE = "Value"


@autocomplete
class Header(Definition):
    pass


@autocomplete
class HeaderRule(DefinitionRule):
    header: Header


@autocomplete
class MediaType(Definition):
    pass


@autocomplete
class MediaTypeRule(DefinitionRule):
    media_type: MediaType


@autocomplete
class HTTPConventionRule(
    ConventionRule[Union[HTTPHeader, HTTPMessage, HTTPMethod, HTTPStatus]]
):
    pass


@autocomplete
class HTTPRuleSet(Ruleset):
    headers: Optional[List[HeaderRule]] = []
    media_types: Optional[List[MediaTypeRule]] = []
    conventions: Optional[List[HTTPConventionRule]] = []
